from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union, Tuple

import transformers as tr

from transformers_embedder import utils

if utils.is_torch_available():
    import torch

logger = utils.get_logger(__name__)
utils.get_logger("transformers")


@dataclass
class TransformersEmbedderOutput(tr.file_utils.ModelOutput):
    """Class for model's outputs."""

    word_embeddings: "torch.Tensor" = None
    last_hidden_state: "torch.FloatTensor" = None
    pooler_output: "torch.FloatTensor" = None
    hidden_states: Optional[Tuple["torch.FloatTensor"]] = None
    attentions: Optional[Tuple["torch.FloatTensor"]] = None


class TransformersEmbedder(torch.nn.Module):
    """
    Transformer Embedder class.
    Word level embeddings from various transformer architectures from Huggingface Trasnformers API.

    Args:
        model (:obj:`str`, :obj:`tr.PreTrainedModel`):
            Transformer model to use (https://huggingface.co/models).
        return_words (:obj:`bool`, optional, defaults to :obj:`True`):
            If ``True`` it returns the word-level embeddings by computing the mean of the
            sub-words embeddings.
        output_layer (:obj:`str`, optional, defaults to :obj:`last`):
            What output to get from the transformer model. The last hidden state (``last``),
            the concatenation of the last four hidden layers (``concat``), the sum of the last four hidden
            layers (``sum``), the pooled output (``pooled``).
        fine_tune (:obj:`bool`, optional, defaults to :obj:`True`):
            If ``True``, the transformer model is fine-tuned during training.
        return_all (:obj:`bool`, optional, defaults to :obj:`False`):
            If ``True``, returns all the outputs from the HuggingFace model.
    """

    def __init__(
        self,
        model: Union[str, tr.PreTrainedModel],
        return_words: bool = True,
        output_layer: str = "last",
        fine_tune: bool = True,
        return_all: bool = False,
    ) -> None:
        super().__init__()
        if isinstance(model, str):
            config = tr.AutoConfig.from_pretrained(model, output_hidden_states=True, output_attentions=True)
            self.transformer_model = tr.AutoModel.from_pretrained(model, config=config)
        else:
            self.transformer_model = model
        self.return_words = return_words
        self.output_layer = output_layer
        self.return_all = return_all
        if not fine_tune:
            for param in self.transformer_model.parameters():
                param.requires_grad = False

    def forward(
        self,
        input_ids: torch.LongTensor,
        attention_mask: torch.BoolTensor = None,
        token_type_ids: Optional[torch.LongTensor] = None,
        offsets: torch.LongTensor = None,
        *args,
        **kwargs,
    ) -> TransformersEmbedderOutput:
        """
        Forward method of the PyTorch module.

        Args:
            input_ids (:obj:`torch.Tensor`, optional):
                Input ids for the transformer model.
            attention_mask (:obj:`torch.Tensor`, optional):
                Attention mask for the transformer model.
            token_type_ids (:obj:`torch.Tensor`, optional):
                Token type ids for the transformer model.
            offsets (:obj:`torch.Tensor`, optional):
                Offsets of the sub-token, used to reconstruct the word embeddings.

        Returns:
             :obj:`TransformersEmbedderOutput`:
                Word level embeddings plus the output of the transformer model.
        """
        # Some of the HuggingFace models don't have the
        # token_type_ids parameter and fail even when it's given as None.
        max_type_id = token_type_ids.max()
        if max_type_id == 0:
            token_type_ids = None
        inputs = {"input_ids": input_ids, "attention_mask": attention_mask}
        if token_type_ids is not None:
            inputs["token_type_ids"] = token_type_ids

        # Shape: [batch_size, num_subtoken, embedding_size].
        transformer_outputs = self.transformer_model(**inputs)
        if self.output_layer == "last":
            word_embeddings = transformer_outputs.last_hidden_state
        elif self.output_layer == "concat":
            word_embeddings = transformer_outputs.hidden_states[-4:]
            word_embeddings = torch.cat(word_embeddings, dim=-1)
        elif self.output_layer == "sum":
            word_embeddings = transformer_outputs.hidden_states[-4:]
            word_embeddings = torch.stack(word_embeddings, dim=0).sum(dim=0)
        elif self.output_layer == "pooled":
            word_embeddings = transformer_outputs.pooler_output
        else:
            raise ValueError(
                "output_layer parameter not valid, choose between `last`, `concat`, "
                f"`sum`, `pooled`. Current value `{self.output_layer}`"
            )

        if self.return_words and offsets is None:
            raise ValueError(
                "`return_words` is `True` but `offsets` was not passed to the model. "
                "Cannot compute word embeddings. To solve:\n"
                "- Set `return_words` to `False` or"
                "- Pass `offsets` to the model during forward."
            )
        if self.return_words:
            # Shape: [batch_size, num_token, embedding_size].
            word_embeddings = self.merge_subword(word_embeddings, offsets)
        if self.return_all:
            return TransformersEmbedderOutput(
                word_embeddings=word_embeddings,
                last_hidden_state=transformer_outputs.last_hidden_state,
                hidden_states=transformer_outputs.hidden_states,
                pooler_output=transformer_outputs.pooler_output,
                attentions=transformer_outputs.attentions,
            )
        return TransformersEmbedderOutput(word_embeddings=word_embeddings)

    def merge_subword(
        self,
        embeddings: torch.Tensor,
        indices: torch.Tensor,
    ) -> torch.Tensor:
        """
        Minimal version of ``scatter_mean``, from `pytorch_scatter <https://github.com/rusty1s/pytorch_scatter/>`_
        library, that is compatible for ONNX but works only for our case. It is used to compute word level
        embeddings from the transformer output.

        Args:
            embeddings (:obj:`torch.Tensor`):
                The embeddings tensor.
            indices (:obj:`torch.Tensor`):
                The subword indices.

        Returns:
            :obj:`torch.Tensor`

        """
        out = self.scatter_sum(embeddings, indices)
        ones = torch.ones(indices.size(), dtype=embeddings.dtype, device=embeddings.device)
        count = self.scatter_sum(ones, indices)
        count.clamp_(1)
        count = self.broadcast(count, out)
        out.true_divide_(count)
        return out

    def scatter_sum(
        self,
        src: torch.Tensor,
        index: torch.Tensor,
    ) -> torch.Tensor:
        """
        Minimal version of ``scatter_sum``, from `pytorch_scatter <https://github.com/rusty1s/pytorch_scatter/>`_
        library, that is compatible for ONNX but works only for our case.

        Args:
            src (:obj:`torch.Tensor`):
                The source tensor.
            index (:obj:`torch.Tensor`):
                The indices of elements to scatter.

        Returns:
            :obj:`torch.Tensor`

        """
        index = self.broadcast(index, src)
        size = list(src.size())
        size[1] = index.max() + 1
        out = torch.zeros(size, dtype=src.dtype, device=src.device)
        return out.scatter_add_(1, index, src)

    def resize_token_embeddings(self, new_num_tokens: Optional[int] = None) -> torch.nn.Embedding:
        """
        Resizes input token embeddings' matrix of the model if :obj:`new_num_tokens != config.vocab_size`.

        Args:
            new_num_tokens (:obj:`int`):
                The number of new tokens in the embedding matrix.

        Returns:
            :obj:`torch.nn.Embedding`: Pointer to the input tokens Embeddings Module of the model.

        """
        return self.transformer_model.resize_token_embeddings(new_num_tokens)

    def save_pretrained(self, save_directory: Union[str, Path]):
        """
        Save a model and its configuration file to a directory.

        Args:
            save_directory (:obj:`str`, :obj:`Path`):
                Directory to which to save.

        Returns:

        """
        self.transformer_model.save_pretrained(save_directory)

    @staticmethod
    def broadcast(src: torch.Tensor, other: torch.Tensor):
        """
        Minimal version of ``broadcast``, from `pytorch_scatter <https://github.com/rusty1s/pytorch_scatter/>`_
        library, that is compatible with ONNX but works only for our case.

        Args:
            src (:obj:`torch.Tensor`):
                The source tensor.
            other (:obj:`torch.Tensor`):
                The tensor from which we want to broadcast.

        Returns:
            :obj:`torch.Tensor`

        """
        for _ in range(src.dim(), other.dim()):
            src = src.unsqueeze(-1)
        src = src.expand_as(other)
        return src

    @property
    def hidden_size(self) -> int:
        """
        Returns the hidden size of the transformer.

        Returns:
            :obj:`int`: Hidden size of ``self.transformer_model``.

        """
        multiplier = 4 if self.output_layer == "concat" else 1
        return self.transformer_model.config.hidden_size * multiplier



import importlib.util
import logging

_torch_available = importlib.util.find_spec("torch") is not None
_spacy_available = importlib.util.find_spec("spacy") is not None


def is_torch_available():
    """
    Check if PyTorch is available.
    """
    return _torch_available


def is_spacy_available():
    """
    Check if spaCy is available.
    """
    return _spacy_available


if is_torch_available():
    import torch
    from torch import Tensor


def get_logger(name: str) -> logging.Logger:
    """
    Return the logger of the given name.
    :param name: name of the logger to return
    :return: the logger
    """
    return logging.getLogger(name)



from transformers_embedder import utils

if utils.is_torch_available():
    from transformers_embedder.embedder import TransformersEmbedder

from transformers import (
    BertTokenizer,
    BertTokenizerFast,
    BertweetTokenizer,
    CamembertTokenizer,
    CamembertTokenizerFast,
    DebertaTokenizer,
    DistilBertTokenizer,
    DistilBertTokenizerFast,
    MobileBertTokenizer,
    MobileBertTokenizerFast,
    RobertaTokenizer,
    RobertaTokenizerFast,
    XLMRobertaTokenizer,
    XLMRobertaTokenizerFast,
    XLMTokenizer,
)


MODELS_WITH_STARTING_TOKEN = (
    BertTokenizer,
    BertTokenizerFast,
    DistilBertTokenizer,
    DistilBertTokenizerFast,
    MobileBertTokenizer,
    MobileBertTokenizerFast,
    BertweetTokenizer,
    CamembertTokenizer,
    CamembertTokenizerFast,
    DebertaTokenizer,
    RobertaTokenizer,
    RobertaTokenizerFast,
    XLMRobertaTokenizer,
    XLMRobertaTokenizerFast,
    XLMTokenizer,
)

MODELS_WITH_DOUBLE_SEP = (
    CamembertTokenizer,
    CamembertTokenizerFast,
    BertweetTokenizer,
    RobertaTokenizer,
    RobertaTokenizerFast,
    XLMRobertaTokenizer,
    XLMRobertaTokenizerFast,
)

from transformers_embedder.tokenizer import Tokenizer



from __future__ import annotations

import math
from collections import UserDict
from functools import partial
from typing import List, Dict, Union, Tuple, Any

import transformers as tr

from transformers_embedder import MODELS_WITH_STARTING_TOKEN, MODELS_WITH_DOUBLE_SEP
from transformers_embedder import utils
from transformers_embedder.utils import is_torch_available, is_spacy_available

if is_torch_available():
    import torch

if is_spacy_available():
    import spacy
    from spacy.cli.download import download as spacy_download

logger = utils.get_logger(__name__)
utils.get_logger("transformers")


class Tokenizer:
    """
    A wrapper class for HuggingFace Tokenizer.

    Args:
        model (:obj:`str`, :obj:`transformers.PreTrainedTokenizer`):
            Language model name (or a transformer :obj:`PreTrainedTokenizer`.
    """

    def __init__(self, model: Union[str, tr.PreTrainedTokenizer], language: str = "xx_sent_ud_sm"):
        if isinstance(model, str):
            # init HuggingFace tokenizer
            self.huggingface_tokenizer = tr.AutoTokenizer.from_pretrained(model)
            # get config
            self.config = tr.AutoConfig.from_pretrained(model)
        else:
            self.huggingface_tokenizer = model
            self.config = tr.AutoConfig.from_pretrained(self.huggingface_tokenizer.name_or_path)
        # simple tokenizer used if the input is `str`
        # lazy load, None at first
        self.spacy_tokenizer = None
        # default multilingual model
        self.language = language
        # padding stuff
        # default, batch length is model max length
        self.subtoken_max_batch_len = self.huggingface_tokenizer.model_max_length
        self.word_max_batch_len = self.huggingface_tokenizer.model_max_length
        # padding ops
        self.padding_ops = {
            "input_ids": partial(
                self.pad_sequence,
                value=self.huggingface_tokenizer.pad_token_id,
                length="subtoken",
            ),
            # value is None because: (read `pad_sequence` doc)
            "offsets": partial(self.pad_sequence, value=None, length="subtoken"),
            "attention_mask": partial(self.pad_sequence, value=0, length="subtoken"),
            "word_mask": partial(self.pad_sequence, value=0, length="word"),
            "token_type_ids": partial(
                self.pad_sequence,
                value=self.token_type_id,
                length="subtoken",
            ),
        }
        # keys that will be converted in tensors
        self.to_tensor_inputs = {
            "input_ids",
            "offsets",
            "attention_mask",
            "word_mask",
            "token_type_ids",
        }

    def __len__(self):
        """Size of the full vocabulary with the added tokens."""
        return len(self.huggingface_tokenizer)

    def __call__(
        self,
        text: Union[str, List[str], List[List[str]]],
        text_pair: Union[str, List[str], List[List[str]]] = None,
        padding: bool = True,
        max_length: int = 0,
        return_tensors: bool = True,
        is_split_into_words: bool = False,
        use_spacy: bool = False,
        *args,
        **kwargs,
    ) -> ModelInputs:
        """
        Prepare the text in input for models that uses HuggingFace as embeddings.

        Args:
            text (:obj:`str`, :obj:`List[str]`, :obj:`List[List[str]]`, :obj:`List[List[Word]]`, :obj:`List[Word]`):
                Text or batch of text to be encoded.
            text_pair (:obj:`str`, :obj:`List[str]`, :obj:`List[List[str]]`, :obj:`List[List[Word]]`, :obj:`List[Word]`):
                Text or batch of text to be encoded.
            padding (:obj:`bool`, optional, defaults to :obj:`True`):
                If :obj:`True`, applies padding to the batch based on the maximum length of the batch.
            max_length (:obj:`int`, optional, defaults to :obj:`0`):
                If specified, truncates the input sequence to that value. Otherwise,
                uses the model max length.
            return_tensors (:obj:`bool`, optional, defaults to :obj:`True`):
                If :obj:`True`, the outputs is converted to :obj:`torch.Tensor`
            is_split_into_words (:obj:`bool`, optional, defaults to :obj:`False`):
                If :obj:`True` and the input is a string, the input is split on spaces.
            use_spacy (:obj:`bool`, optional, defaults to :obj:`False`):
                If :obj:`True` the input is split using SpaCy, otherwise it is split on spaces.

        Returns:
            :obj:`ModelInputs`: The inputs to the transformer model.

        """
        # type checking before everything
        # self._type_checking(text, text_pair)

        # check if input is batched or a single sample
        is_batched = bool(isinstance(text, (list, tuple)) and text and isinstance(text[0], (list, tuple)))

        # if text is str or a list of str and they are not split, then text needs to be tokenized
        if isinstance(text, str) or (not is_split_into_words and isinstance(text[0], str)):
            if not is_batched:
                text = self.pretokenize(text, use_spacy=use_spacy)
                text_pair = self.pretokenize(text_pair, use_spacy=use_spacy) if text_pair else None
            else:
                text = [self.pretokenize(t, use_spacy=use_spacy) for t in text]
                text_pair = (
                    [self.pretokenize(t, use_spacy=use_spacy) for t in text_pair] if text_pair else None
                )

        # get model max length if not specified by user
        if max_length == 0:
            max_length = self.huggingface_tokenizer.model_max_length

        if not is_batched:
            output = self.build_tokens(text, text_pair, max_length)
        else:
            if not padding and return_tensors:
                logger.info(
                    "`padding` is `False` and return_tensor is `True`. Cannot make tensors from "
                    "not padded sequences. `padding` forced automatically to `True`"
                )
                padding = True
            output = self.build_tokens_batch(text, text_pair, max_length)

        # clean the output
        output = self._clean_output(output)
        if padding:
            output = self.pad_batch(output)
        if return_tensors:
            output = self.to_tensor(output)
        output = ModelInputs(output)
        return output

    def build_tokens_batch(
        self,
        text: List[List[str]],
        text_pair: List[List[str]] = None,
        max_length: int = math.inf,
    ) -> List[Dict[str, Union[list, int]]]:
        """
        Builds the batched input.

        Args:
            text (:obj:`List[List[Word]]`, :obj:`List[Word]`):
                Text or batch of text to be encoded.
            text_pair (:obj:`List[List[Word]]`, :obj:`List[Word]`):
                Text pair or batch of text to be encoded.
           max_length (:obj:`int`, optional, defaults to :obj:`0`):
                If specified, truncates the input sequence to that value. Otherwise,
                uses the model max length.

        Returns:
            :obj:`List[Dict[str, Union[list, int]]]`: The encoded batch

        """
        batch = []
        if not text_pair:
            # In this way we can re-use the already defined methods,
            # regardless the presence of the text pairs
            text_pair = [None for _ in text]
        for t, t_p in zip(text, text_pair):
            token_pair = self.build_tokens(t, t_p, max_length)
            batch.append(token_pair)
        return batch

    def build_tokens(
        self, text: List[str], text_pair: List[str] = None, max_length: int = math.inf
    ) -> Dict[str, Union[list, int]]:
        """
        Build transformer pair input.

        Args:
            text (:obj:`List[Word]`):
                Text to be encoded.
            text_pair (:obj:`List[Word]`):
                Text pair to be encoded.
           max_length (:obj:`int`, optional, defaults to :obj:`0`):
                If specified, truncates the input sequence to that value. Otherwise,
                uses the model max length.

        Returns:
            :obj:`Dict[str, Union[list, int]]`: A dictionary with :obj:`text` and :obj:`text_pair` encoded.
        """
        words, input_ids, token_type_ids, offsets = self._build_tokens(text, max_length=max_length)
        if text_pair:
            words_b, input_ids_b, token_type_ids_b, offsets_b = self._build_tokens(
                text_pair, True, max_length
            )
            # align offsets of sentence b
            offsets_b = [o + len(words) for o in offsets_b]
            offsets = offsets + offsets_b
            input_ids += input_ids_b
            token_type_ids += token_type_ids_b
            words += words_b

        word_mask = [1] * len(words)  # for original tokens
        attention_mask = [1] * len(input_ids)

        return {
            "words": words,
            "input_ids": input_ids,
            "offsets": offsets,
            "attention_mask": attention_mask,
            "word_mask": word_mask,
            "token_type_ids": token_type_ids,
            "sentence_lengths": len(words),
        }

    def _build_tokens(
        self, text: List[str], is_b: bool = False, max_length: int = math.inf
    ) -> Tuple[list, list, list, list]:
        """
        Encode the sentence for transformer model.

        Args:
            text (:obj:`List[Word]`):
                Text to encode.
            is_b (:obj:`bool`, optional, defaults to :obj:`False`):
                if :obj:`True`, skips first :obj:`CLS` token and set token_type_id to :obj:`1`
            max_length (:obj:`int`, optional, defaults to :obj:`0`):
                If specified, truncates the input sequence to that value. Otherwise,
                uses the model max length.

        Returns:
            :obj:`Tuple`:
                The encoded sentence, with :obj:`input_ids`, :obj:`token_type_ids` and :obj:`offsets`.
        """
        # words
        words = []
        # HuggingFace model inputs
        input_ids = []
        token_type_ids = []
        offsets = []
        offset_starting_index = 0
        if not is_b:
            token_type_id = 0
            # some models don't need starting special token
            if isinstance(self.huggingface_tokenizer, MODELS_WITH_STARTING_TOKEN):
                words += [self.huggingface_tokenizer.cls_token]
                input_ids += [self.huggingface_tokenizer.cls_token_id]
                token_type_ids += [token_type_id]
                # first offset
                offsets.append(0)
                offset_starting_index = 1
        else:
            token_type_id = self.token_type_id
            # check if the input needs an additional sep token
            # XLM-R for example wants an additional `</s>` between text pairs
            if isinstance(self.huggingface_tokenizer, MODELS_WITH_DOUBLE_SEP):
                words += [self.huggingface_tokenizer.sep_token]
                input_ids += [self.huggingface_tokenizer.sep_token_id]
                token_type_ids += [token_type_id]
                offsets.append(0)
                offset_starting_index = 1

        for word_index, word in enumerate(text):
            ids = self.huggingface_tokenizer(word, add_special_tokens=False)["input_ids"]
            # if max_len exceeded, stop (leave space for closing token)
            if len(input_ids) + len(ids) >= max_length - 1:
                break
            # token offset before wordpiece, (start, end + 1)
            # offsets.append((len(input_ids), len(input_ids) + len(ids) - 1))
            offsets += [word_index + offset_starting_index] * len(ids)
            words += [word]
            input_ids += ids
            token_type_ids += [token_type_id] * len(ids)
        # last offset
        words += [self.huggingface_tokenizer.sep_token]
        input_ids += [self.huggingface_tokenizer.sep_token_id]
        token_type_ids += [token_type_id]
        offsets.append(len(words) - 1)  # -1 because we want the last index
        return words, input_ids, token_type_ids, offsets

    def pad_batch(self, batch: Union[ModelInputs, Dict[str, list]], max_length: int = None) -> ModelInputs:
        """
        Pad the batch to its maximum length or to the specified :obj:`max_length`.

        Args:
            batch (:obj:`Dict[str, list]`):
                The batch to pad.
            max_length (:obj:`int`, optional):
                Override maximum length of the batch.

        Returns:
            :obj:`Dict[str, list]`: The padded batch.
        """
        if max_length:
            self.subtoken_max_batch_len = max_length
            self.word_max_batch_len = max_length
        else:
            # get maximum len inside a batch
            self.subtoken_max_batch_len = max(len(x) for x in batch["input_ids"])
            self.word_max_batch_len = max(x for x in batch["sentence_lengths"])

        for key in batch:
            if key in self.padding_ops:
                batch[key] = [self.padding_ops[key](b) for b in batch[key]]
        return ModelInputs(batch)

    def pad_sequence(
        self,
        sequence: Union[List, torch.Tensor],
        value: Any = None,
        length: Union[int, str] = "subtoken",
        pad_to_left: bool = False,
    ) -> Union[List, torch.Tensor]:
        """
        Pad the input to the specified length with the given value.

        Args:
            sequence (:obj:`List`, :obj:`torch.Tensor`):
                Element to pad, it can be either a :obj:`List` or a :obj:`torch.Tensor`.
            value (:obj:`Any`, optional):
                Value to use as padding.
            length (:obj:`int`, :obj:`str`, optional, defaults to :obj:`subtoken`):
                Length after pad.
            pad_to_left (:obj:`bool`, optional, defaults to :obj:`False`):
                If :obj:`True`, pads to the left, right otherwise.

        Returns:
            :obj:`List`, :obj:`torch.Tensor`: The padded sequence.

        """
        if length == "subtoken":
            length = self.subtoken_max_batch_len
        elif length == "word":
            length = self.word_max_batch_len
        else:
            if not isinstance(length, int):
                raise ValueError(
                    f"`length` must be an `int`, `subtoken` or `word`. Current value is `{length}`"
                )
        if value is None:
            # this is a trick used to pad the offset.
            # here we want the offset pad to be the max offset index in the batch
            # that is the maximum word length in the batch
            value = self.word_max_batch_len - 1
        padding = [value] * abs(length - len(sequence))
        if isinstance(sequence, torch.Tensor):
            if len(sequence.shape) > 1:
                raise ValueError(f"Sequence tensor must be 1D. Current shape is `{len(sequence.shape)}`")
            padding = torch.as_tensor(padding)
        if pad_to_left:
            if isinstance(sequence, torch.Tensor):
                return torch.cat((padding, sequence), -1)
            return padding + sequence
        if isinstance(sequence, torch.Tensor):
            return torch.cat((sequence, padding), -1)
        return sequence + padding

    def pretokenize(self, text: str, use_spacy: bool = False) -> List[str]:
        """
        Pre-tokenize the text in input, splitting on spaces or using SpaCy tokenizer if `use_spacy` is True.

        Args:
            text (:obj:`str`):
                The text to pre-tokenize.
            use_spacy (:obj:`bool`, optional, defaults to :obj:`False`):
                If :obj:`True`, uses SpaCy tokenizer.

        Returns:
            :obj:`List[str]`: The pre-tokenized text.
        """
        if use_spacy:
            if not self.spacy_tokenizer:
                self._load_spacy()
            text = self.spacy_tokenizer(text)
            return [t.text for t in text]
        return text.split(" ")

    def add_special_tokens(self, special_tokens_dict: Dict[str, Union[str, tr.AddedToken]]) -> int:
        """
        Add a dictionary of special tokens (eos, pad, cls, etc.) to the encoder.
        If special tokens are NOT in the vocabulary, they are added to it (indexed starting from the last
        index of the current vocabulary).

        Args:
            special_tokens_dict (:obj:`Dict`):
                The dictionary containing special tokens. Keys should be in
                the list of predefined special attributes: [``bos_token``, ``eos_token``,
                ``unk_token``, ``sep_token``, ``pad_token``, ``cls_token``, ``mask_token``,
                ``additional_special_tokens``].

        Returns:
            :obj:`int`: Number of tokens added to the vocabulary.

        """
        return self.huggingface_tokenizer.add_special_tokens(special_tokens_dict)

    def add_padding_ops(self, key: str, value: Any, length: Union[int, str]):
        """
        Add padding logic to custom fields.
        If the field is not in :obj:`self.to_tensor_inputs`, this method will add the key to it.

        Args:
            key (:obj:`str`):
                Name of the field in the tokenizer input.
            value (:obj:`Any`):
                Value to use for padding.
            length (:obj:`int`, :obj:`str`):
                Length to pad. It can be an :obj:`int`, or two string value
                - :obj:`subtoken`: the element is padded to the batch max length relative to the subtokens length
                - :obj:`word`: the element is padded to the batch max length relative to the original word length

        Returns:

        """
        if key not in self.to_tensor_inputs:
            self.to_tensor_inputs.add(key)
        self.padding_ops[key] = partial(self.pad_sequence, value=value, length=length)

    def add_to_tensor_inputs(self, names: Union[str, set]):
        """
        Add these keys to the ones that will be converted in Tensors.

        Args:
            names (:obj:`str`, :obj:`set`):
                Name of the field (or fields) to convert to tensors.

        Returns:

        """
        if isinstance(names, str):
            names = {names}
        self.to_tensor_inputs |= names

    def to_tensor(self, batch: Union[ModelInputs, List[dict], dict]) -> ModelInputs:
        """
        Return a the batch in input as Pytorch tensors.
        The fields that are converted in tensors are in :obj:`self.to_tensor_inputs`. By default, only the
        standard model inputs are converted. Use :obj:`self.add_to_tensor_inputs` to add custom fields.

        Args:
            batch (:obj:`List[dict]`, :obj:`dict`):
                Batch in input.

        Returns:
            :obj:`dict`: The batch as tensor.

        """
        # convert to tensor
        batch = {
            k: torch.as_tensor(v) if k in self.to_tensor_inputs and not isinstance(v, torch.Tensor) else v
            for k, v in batch.items()
        }
        return ModelInputs(batch)

    def _load_spacy(self) -> "spacy.tokenizer.Tokenizer":
        """
        Download and load spacy model.

        Returns:
            :obj:`spacy.tokenizer.Tokenizer`: The spacy tokenizer.
        """
        try:
            spacy_tagger = spacy.load(self.language, exclude=["ner", "parser"])
        except OSError:
            logger.info(f"Spacy model '{self.language}' not found. Downloading and installing.")
            spacy_download(self.language)
            spacy_tagger = spacy.load(self.language, exclude=["ner", "parser"])
        self.spacy_tokenizer = spacy_tagger.tokenizer
        return self.spacy_tokenizer

    @staticmethod
    def _clean_output(output: Union[List, Dict]) -> Dict:
        """
        Clean before output.

        Args:
            output (:obj`List[dict]`, :obj:`dict`):
                The output to clean.

        Returns:
            :obj:`dict`: The cleaned output.

        """
        # single sentence case, generalize
        if isinstance(output, dict):
            output = [output]
        # convert list to dict
        output = {k: [d[k] for d in output] for k in output[0]}
        return output

    @staticmethod
    def _get_token_type_id(config: tr.PretrainedConfig) -> int:
        """
        Get token type id. Useful when dealing with models that don't accept 1 as type id.
        Args:
            config (:obj:`transformers.PretrainedConfig`):
                Transformer config.

        Returns:
            :obj:`int`: Correct token type id for that model.

        """
        if hasattr(config, "type_vocab_size"):
            return 1 if config.type_vocab_size == 2 else 0
        return 0

    @staticmethod
    def _type_checking(text: Any, text_pair: Any):
        """
        Checks type of the inputs.

        Args:
            text (:obj:`Any`):
                Text to check.
            text_pair (:obj:`Any`):
                Text pair to check.

        Returns:

        """

        def is_type_correct(text_to_check: Any) -> bool:
            """
            Check if input type is correct, returning a boolean.

            Args:
                text_to_check (:obj:`Any`):
                    text to check.

            Returns:
                :obj`bool`: :obj`True` if the type is correct.

            """
            return (
                text_to_check is None
                or isinstance(text_to_check, str)
                or (
                    isinstance(text_to_check, (list, tuple))
                    and (
                        len(text_to_check) == 0
                        or (
                            isinstance(text_to_check[0], str)
                            or (
                                isinstance(text_to_check[0], (list, tuple))
                                and (len(text_to_check[0]) == 0 or isinstance(text_to_check[0][0], str))
                            )
                        )
                    )
                )
            )

        if not is_type_correct(text):
            raise AssertionError(
                "text input must of type `str` (single example), `List[str]` (batch or single "
                "pretokenized example) or `List[List[str]]` (batch of pretokenized examples)."
            )

        if not is_type_correct(text_pair):
            raise AssertionError(
                "text_pair input must be `str` (single example), `List[str]` (batch or single "
                "pretokenized example) or `List[List[str]]` (batch of pretokenized examples)."
            )

    @property
    def num_special_tokens(self) -> int:
        """
        Return the number of special tokens the model needs.
        It assume the input contains both sentences (:obj:`text` and :obj:`text_pair`).

        Returns:
            :obj:`int`: the number of special tokens.

        """
        if isinstance(self.huggingface_tokenizer, MODELS_WITH_DOUBLE_SEP) and isinstance(
            self.huggingface_tokenizer, MODELS_WITH_STARTING_TOKEN
        ):
            return 4
        if isinstance(
            self.huggingface_tokenizer,
            (MODELS_WITH_DOUBLE_SEP, MODELS_WITH_STARTING_TOKEN),
        ):
            return 3
        return 2

    @property
    def has_double_sep(self):
        """True if tokenizer uses two SEP tokens."""
        return isinstance(self.huggingface_tokenizer, MODELS_WITH_DOUBLE_SEP)

    @property
    def has_starting_token(self):
        """True if tokenizer uses a starting token."""
        return isinstance(self.huggingface_tokenizer, MODELS_WITH_STARTING_TOKEN)

    @property
    def token_type_id(self):
        """Padding token."""
        return self._get_token_type_id(self.config)

    @property
    def pad_token(self):
        """Padding token."""
        return self.huggingface_tokenizer.pad_token

    @property
    def pad_token_id(self):
        """Padding token id."""
        return self.huggingface_tokenizer.pad_token_id

    @property
    def unk_token(self):
        """Unknown token."""
        return self.huggingface_tokenizer.unk_token

    @property
    def unk_token_id(self):
        """Unknown token id."""
        return self.huggingface_tokenizer.unk_token_id

    @property
    def cls_token(self):
        """
        Classification token.
        To extract a summary of an input sequence leveraging self-attention along the
        full depth of the model.
        """
        return self.huggingface_tokenizer.cls_token

    @property
    def cls_token_id(self):
        """
        Classification token id.
        To extract a summary of an input sequence leveraging self-attention along the
        full depth of the model.
        """
        return self.huggingface_tokenizer.cls_token_id

    @property
    def sep_token(self):
        """Separation token, to separate context and query in an input sequence."""
        return self.huggingface_tokenizer.sep_token

    @property
    def sep_token_id(self):
        """Separation token id, to separate context and query in an input sequence."""
        return self.huggingface_tokenizer.sep_token_id

    @property
    def bos_token(self):
        """Beginning of sentence token."""
        return self.huggingface_tokenizer.bos_token

    @property
    def bos_token_id(self):
        """Beginning of sentence token id."""
        return self.huggingface_tokenizer.bos_token_id

    @property
    def eos_token(self):
        """End of sentence token."""
        return self.huggingface_tokenizer.eos_token

    @property
    def eos_token_id(self):
        """End of sentence token id."""
        return self.huggingface_tokenizer.eos_token_id


class ModelInputs(UserDict):
    """Model input dictionary wrapper."""

    def __getattr__(self, item: str):
        try:
            return self.data[item]
        except KeyError:
            raise AttributeError

    def __getitem__(self, item: str) -> Any:
        return self.data[item]

    def __getstate__(self):
        return {"data": self.data}

    def __setstate__(self, state):
        if "data" in state:
            self.data = state["data"]

    def keys(self):
        """A set-like object providing a view on D's keys."""
        return self.data.keys()

    def values(self):
        """An object providing a view on D's values."""
        return self.data.values()

    def items(self):
        """A set-like object providing a view on D's items."""
        return self.data.items()

    def to(self, device: Union[str, torch.device]) -> ModelInputs:
        """
        Send all tensors values to device.

        Args:
            device (:obj:`str` or :obj:`torch.device`): The device to put the tensors on.

        Returns:
            :class:`tokenizers.ModelInputs`: The same instance of
        :class:`~tokenizers.ModelInputs` after modification.

        """
        if isinstance(device, (str, torch.device, int)):
            self.data = {
                k: v.to(device=device) if isinstance(v, torch.Tensor) else v for k, v in self.data.items()
            }
        else:
            logger.warning(f"Attempting to cast to another type, {str(device)}. This is not supported.")
        return self



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

extras = {}
extras["torch"] = ["torch>=1.5,<1.11"]
extras["spacy"] = ["spacy>=3.0,<3.3"]
extras["all"] = extras["torch"] + extras["spacy"]

install_requires = ["transformers>=4.3,<4.17"]

setuptools.setup(
    name="transformers_embedder",  # Replace with your own username
    version="1.8.5",
    author="Riccardo Orlando",
    author_email="orlandoricc@gmail.com",
    description="Word level transformer based embeddings",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Riccorl/transformers-embedder",
    keywords="NLP deep learning transformer pytorch BERT google subtoken wordpieces embeddings",
    packages=setuptools.find_packages(),
    include_package_data=True,
    license="Apache",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    extras_require=extras,
    install_requires=install_requires,
    python_requires=">=3.6",
)
